#include "EXTERN.h"
#include "perl.h"
#include "nexus.h"
#include "pnx.h"

/*
 * Nexus attachment support.
 *
 * Since we can't directly invoke the allow_attach, we set up
 * a table mapping from url to gp.
 *
 * This table is manipulated with the calls
 *
 *    Nexus::set_attachment($url, $ref)
 *    Nexus::remove_attachment($url)
 *    Nexus::set_default_attachment($ref)
 *
 * The table is kept currently as a simple list mapping from
 * URL to its global pointer. The perl-side calls create a RRT
 * object to capture the reference, and store the RRT's global pointer
 * in the table.
 *
 * When the RRT is created for a table entry, the RRT refcount is
 * incremented (since there is the "potential" for creation of a
 * remote reference while an attachment is allowed). When the
 * entry is removed, we decrement the refcount (and dont worry about
 * destroying the object).
 *
 * Note also we need to lock around attachment table operations, as
 * an attachment request can come in at any point (it is handled
 * by the network handler thread).
 *
 * The call Nexus::allow_attach($port) sets up a Nexus attachment listener.
 * It returns three element list: ($rc, $port, $host). $rc is the return
 * code from the Nexus call, zero indicating success. $port and $host
 * specify the port and host needed for another process to connect.
 *
 * The call Nexus::attach_with_url($url) attaches to another Nexus program
 * with the given url. If it is successful, it will return the
 * Nexus::GlobalPointer as returned from the serving Nexus program.
 *
 * Note that if we're talking to another nperl program, we will have
 * to convert the global pointer into a RemoteReference. This is done
 * at nperl library level, not here. This allows for communication of nperl
 * programs with other Nexus programs.
 *
 */

/*
 * Key returned for the default entry for get_attachment_table();
 */
#define DEFAULT_KEY "DEFAULT"

typedef struct _attachment_table_entry_t
{
    char *url;
    SV *rrt;
    nexus_global_pointer_t gp;
    struct _attachment_table_entry_t *next;
} attachment_table_entry_t;

typedef struct _attachment_table_t
{
    attachment_table_entry_t entries;
    attachment_table_entry_t default_entry;
    nexus_mutex_t lock;
} attachment_table_t;

static attachment_table_t attachment_table;

static void lookup_entry(char *url, attachment_table_entry_t **entryp,
			 attachment_table_entry_t **prevp);
static void insert_entry(attachment_table_entry_t *entry);
static void invalidate_entry(attachment_table_entry_t *entry);
static void setup_entry(attachment_table_entry_t *entry, char *url, SV *ref);

void pnx_attach_boot()
{
    attachment_table.entries.next = (attachment_table_entry_t *) NULL;

    nexus_mutex_init(&attachment_table.lock, (nexus_mutexattr_t *) NULL);
    
    nexus_null_global_pointer(&attachment_table.default_entry.gp);
    attachment_table.default_entry.rrt = (SV *) NULL;
    attachment_table.default_entry.url = (char *) NULL;
}

void pnx_set_attachment(char *url, SV *ref)
{
    attachment_table_entry_t *entry;

    nexus_mutex_lock(&attachment_table.lock);

    lookup_entry(url, &entry, NULL);
    if (entry == (attachment_table_entry_t *) NULL)
    {
	NexusMalloc(pnx_set_attachment, entry, attachment_table_entry_t *,
		    sizeof(attachment_table_entry_t));
	insert_entry(entry);
    }
    else
    {
	invalidate_entry(entry);
    }

    setup_entry(entry, url, ref);

    nexus_mutex_unlock(&attachment_table.lock);
}

void pnx_remove_attachment(char *url)
{
    attachment_table_entry_t *entry, *prev;

    nexus_mutex_lock(&attachment_table.lock);

    lookup_entry(url, &entry, &prev);
    if (entry == (attachment_table_entry_t *) NULL)
    {
	nexus_mutex_unlock(&attachment_table.lock);
	return;
    }

    invalidate_entry(entry);
    prev->next = entry->next;

    NexusFree(entry);
    nexus_mutex_unlock(&attachment_table.lock);
}

void pnx_set_default_attachment(SV *ref)
{
    nexus_mutex_lock(&attachment_table.lock);

    invalidate_entry(&attachment_table.default_entry);
    setup_entry(&attachment_table.default_entry, NULL, ref);
    nexus_mutex_unlock(&attachment_table.lock);
}

void pnx_remove_default_attachment()
{
    nexus_mutex_lock(&attachment_table.lock);

    invalidate_entry(&attachment_table.default_entry);

    nexus_mutex_unlock(&attachment_table.lock);
}

HV *pnx_get_attachment_table()
{
    attachment_table_entry_t *entry;
    HV *hash;

    nexus_mutex_lock(&attachment_table.lock);

    hash = newHV();

    for (entry = attachment_table.entries.next;
	 entry != (attachment_table_entry_t *) NULL;
	 entry = entry->next)
    {
	hv_store(hash, entry->url, strlen(entry->url), entry->rrt, 0);
    }

    if (attachment_table.default_entry.rrt != (SV *) NULL)
    {
	hv_store(hash, DEFAULT_KEY, strlen(DEFAULT_KEY),
		 attachment_table.default_entry.rrt, 0);
    }
    
    nexus_mutex_unlock(&attachment_table.lock);

    return hash;
}

static int approve_attach(void *arg, char *url, nexus_global_pointer_t *gp)
{
    attachment_table_entry_t *entry, *def;
    int rc;
    SV *func;

    func = (SV *) arg;

    /*
     * If we have a perl approval function, invoke that. Otherwise
     * do the table lookup.
     *
     * The perl function doesn't get to have an argument yet. If it
     * really wants one it can be done with an anonymous subroutine
     * and a lexical var.
     *
     * Arguments are the url and reference to a global pointer.
     *
     */

    if (func != &sv_undef)
    {
	dSP;
	int count, rc;
	SV *gpref, *gpvar;

	gpref = newRV(newSVsv(&sv_undef));
/*	printf("gpref=%x gpv=%x\n", gpref, SvRV(gpref));*/

	ENTER;
	SAVETMPS;
	PUSHMARK(sp);
	XPUSHs(sv_2mortal(newSVpv(url, strlen(url))));
	XPUSHs(gpref);
	PUTBACK ;
	
	count = perl_call_sv(func, G_EVAL | G_SCALAR);
        SPAGAIN ;
	if (count != 1)
	    croak("perl_call_sv shoulda returned one");

	rc = POPi;
	
	PUTBACK ;
	FREETMPS ;
	LEAVE ;

	/*
        printf("call_sv returns %d\n", rc);
	printf("gpref=%x gpv=%x\n", gpref, SvRV(gpref));
	*/

	if (rc == 0)
	{
	    /*
	    printf("returned sv:\n");
	    pnx_sv_dump(gpref);
	    */

	    gpvar = SvRV(gpref);

	    if (sv_isa(gpvar, "Nexus::GlobalPointer"))
	    {
		/*
		printf("got gp\n");
		*/
		*gp = *(pnx_gp_get_gp(gpvar, "approve_attach"));
	    }
	    else
	    {
		SV *rrt, *gpsv;

		/*
		printf("got other type\n");
		*/
		
		rrt = pnx_rrt_new(gpvar);
		gpsv = pnx_rrt_gp(rrt);
		*gp = *(pnx_gp_get_gp(gpsv, "approve_attach"));
		
		pnx_rrt_incr(rrt, 1);
	    }
	    /*
	    printf("gp is %s\n", nexus_global_pointer_string(gp));
	    */
	}

	return rc;
    }


    nexus_mutex_lock(&attachment_table.lock);

/*    printf("Got attachment request for %s\n", url);*/

    lookup_entry(url, &entry, (attachment_table_entry_t **) NULL);

    rc = 1;
    if (entry)
    {
/*	printf("Approved from table entry\n");*/
	*gp = entry->gp;
	rc = 0;
    }
    else
    {
	def = &attachment_table.default_entry;
	if (def->rrt != (SV *) NULL)
	{
/*	    printf("Approved from default\n");*/
	    *gp = def->gp;
	    rc = 0;
	}
    }
    
    nexus_mutex_unlock(&attachment_table.lock);

    return rc;
}

int pnx_allow_attach(unsigned short *port, char **host, SV *func)
{
    int rc;

    if (func != &sv_undef)
    {
	SvREFCNT_inc(func);
    }

    rc = nexus_allow_attach(port, host, approve_attach, func);

    return rc;
}

void pnx_disallow_attach(int port)
{
    nexus_disallow_attach((unsigned short) port);
}

SV *pnx_attach(char *url)
{
    nexus_global_pointer_t *gp;
    int rc;
    SV *gpsv;

    NexusMalloc(pnx_attach, gp, nexus_global_pointer_t *,
		sizeof(nexus_global_pointer_t));
    
    rc = nexus_attach(url, gp);
/*    printf("attach_with_url returns %d\n", rc);*/
    if (rc == 0)
    {
	gpsv = sv_setref_iv(newSVsv(&sv_undef),
			     "Nexus::GlobalPointer", (IV) gp);
    }
    else
    {
	SV *sv;
	NexusFree(gp);
	gpsv = newSVsv(&sv_undef);

	sv = perl_get_sv("Nexus::error", TRUE);
	sv_setiv(sv, (IV) rc);
	sv_setpv(sv, nexus_fault_strings[rc]);
	SvIOK_on(sv);
    }
    return gpsv;
}

static void lookup_entry(char *url, attachment_table_entry_t **entryp,
			 attachment_table_entry_t **prevp)
{
    attachment_table_entry_t *entry, *prev;

    prev = &attachment_table.entries;
    entry = prev->next;
    while (entry != (attachment_table_entry_t *) NULL)
    {
	if (strcmp(entry->url, url) == 0)
	{
	    *entryp = entry;
	    if (prevp != (attachment_table_entry_t **) NULL)
		*prevp = prev;
	    return;
	}
	prev = entry;
	entry = entry->next;
    }
    *entryp = (attachment_table_entry_t *) NULL;
    if (prevp != (attachment_table_entry_t **) NULL)
	*prevp = (attachment_table_entry_t *) NULL;
}

static void insert_entry(attachment_table_entry_t *entry)
{
    entry->next = attachment_table.entries.next;
    attachment_table.entries.next = entry;
}

static void invalidate_entry(attachment_table_entry_t *entry)
{
    if (entry->rrt != (SV *) NULL)
	pnx_rrt_decr(entry->rrt, 1);
    entry->rrt = (SV *) NULL;
    
    if (entry->url != (char *) NULL)
	NexusFree(entry->url);
    entry->url = (char *) NULL;
    nexus_null_global_pointer(&entry->gp);
}

static void setup_entry(attachment_table_entry_t *entry, char *url, SV *ref)
{
    SV *gpsv;
    nexus_global_pointer_t *gp;

    if (url == (char *) NULL)
    {
	entry->url = (char *) NULL;
    }
    else
    {
	NexusMalloc(setup_entry, entry->url, char *,
		    strlen(url) + 1);
	strcpy(entry->url, url);
    }

    entry->rrt = pnx_rrt_new(ref);
    gpsv = pnx_rrt_gp(entry->rrt);
    gp = pnx_gp_get_gp(gpsv, "setup_entry");
    entry->gp = *gp;

    pnx_rrt_incr(entry->rrt, 1);
}
